@@ -13,24 +13,31 @@ end of the VS buffer table.
 
 !*/
 
+// `MTLFeatureSet` is superseded by `MTLGpuFamily`.
+// However, `MTLGpuFamily` is only supported starting MacOS 10.15, whereas our minimum target is MacOS 10.13,
+// See https://github.com/gpuweb/gpuweb/issues/1069 for minimum spec.
+// TODO: Eventually all deprecated features should be abstracted and use new api when available.
+#[allow(deprecated)]
 mod adapter;
 mod command;
 mod conv;
 mod device;
 mod surface;
+mod time;
 
 use std::{
-    iter, ops,
+    fmt, iter, ops,
     ptr::NonNull,
     sync::{atomic, Arc},
     thread,
 };
 
 use arrayvec::ArrayVec;
-use foreign_types::ForeignTypeRef as _;
+use bitflags::bitflags;
+use metal::foreign_types::ForeignTypeRef as _;
 use parking_lot::Mutex;
 
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct Api;
 
 type ResourceIndex = u32;
@@ -66,14 +73,15 @@ pub struct Instance {
 }
 
 impl Instance {
-    pub fn create_surface_from_layer(&self, layer: &mtl::MetalLayerRef) -> Surface {
+    pub fn create_surface_from_layer(&self, layer: &metal::MetalLayerRef) -> Surface {
         unsafe { Surface::from_layer(layer) }
     }
 }
 
 impl crate::Instance<Api> for Instance {
     unsafe fn init(_desc: &crate::InstanceDescriptor) -> Result<Self, crate::InstanceError> {
-        //TODO: enable `METAL_DEVICE_WRAPPER_TYPE` environment based on the flags?
+        // We do not enable metal validation based on the validation flags as it affects the entire
+        // process. Instead, we enable the validation inside the test harness itself in tests/src/native.rs.
         Ok(Instance {
             managed_metal_layer_delegate: surface::HalManagedMetalLayerDelegate::new(),
         })
@@ -81,29 +89,31 @@ impl crate::Instance<Api> for Instance {
 
     unsafe fn create_surface(
         &self,
-        has_handle: &impl raw_window_handle::HasRawWindowHandle,
+        _display_handle: raw_window_handle::RawDisplayHandle,
+        window_handle: raw_window_handle::RawWindowHandle,
     ) -> Result<Surface, crate::InstanceError> {
-        match has_handle.raw_window_handle() {
+        match window_handle {
             #[cfg(target_os = "ios")]
             raw_window_handle::RawWindowHandle::UiKit(handle) => {
                 let _ = &self.managed_metal_layer_delegate;
-                Ok(Surface::from_uiview(handle.ui_view))
+                Ok(unsafe { Surface::from_view(handle.ui_view, None) })
             }
             #[cfg(target_os = "macos")]
-            raw_window_handle::RawWindowHandle::AppKit(handle) => Ok(Surface::from_nsview(
-                handle.ns_view,
-                &self.managed_metal_layer_delegate,
-            )),
-            _ => Err(crate::InstanceError),
+            raw_window_handle::RawWindowHandle::AppKit(handle) => Ok(unsafe {
+                Surface::from_view(handle.ns_view, Some(&self.managed_metal_layer_delegate))
+            }),
+            _ => Err(crate::InstanceError::new(format!(
+                "window handle {window_handle:?} is not a Metal-compatible handle"
+            ))),
         }
     }
 
     unsafe fn destroy_surface(&self, surface: Surface) {
-        surface.dispose();
+        unsafe { surface.dispose() };
     }
 
     unsafe fn enumerate_adapters(&self) -> Vec<crate::ExposedAdapter<Api>> {
-        let devices = mtl::Device::all();
+        let devices = metal::Device::all();
         let mut adapters: Vec<crate::ExposedAdapter<Api>> = devices
             .into_iter()
             .map(|dev| {
@@ -114,11 +124,9 @@ impl crate::Instance<Api> for Instance {
                         name,
                         vendor: 0,
                         device: 0,
-                        device_type: if shared.private_caps.low_power {
-                            wgt::DeviceType::IntegratedGpu
-                        } else {
-                            wgt::DeviceType::DiscreteGpu
-                        },
+                        device_type: shared.private_caps.device_type(),
+                        driver: String::new(),
+                        driver_info: String::new(),
                         backend: wgt::Backend::Metal,
                     },
                     features: shared.private_caps.features(),
@@ -137,19 +145,39 @@ impl crate::Instance<Api> for Instance {
     }
 }
 
+bitflags!(
+    /// Similar to `MTLCounterSamplingPoint`, but a bit higher abstracted for our purposes.
+    #[derive(Debug, Copy, Clone)]
+    pub struct TimestampQuerySupport: u32 {
+        /// On creating Metal encoders.
+        const STAGE_BOUNDARIES = 1 << 1;
+        /// Within existing draw encoders.
+        const ON_RENDER_ENCODER = Self::STAGE_BOUNDARIES.bits() | (1 << 2);
+        /// Within existing dispatch encoders.
+        const ON_COMPUTE_ENCODER = Self::STAGE_BOUNDARIES.bits() | (1 << 3);
+        /// Within existing blit encoders.
+        const ON_BLIT_ENCODER = Self::STAGE_BOUNDARIES.bits() | (1 << 4);
+
+        /// Within any wgpu render/compute pass.
+        const INSIDE_WGPU_PASSES = Self::ON_RENDER_ENCODER.bits() | Self::ON_COMPUTE_ENCODER.bits();
+    }
+);
+
 #[allow(dead_code)]
 #[derive(Clone, Debug)]
 struct PrivateCapabilities {
     family_check: bool,
-    msl_version: mtl::MTLLanguageVersion,
+    msl_version: metal::MTLLanguageVersion,
     fragment_rw_storage: bool,
-    read_write_texture_tier: mtl::MTLReadWriteTextureTier,
+    read_write_texture_tier: metal::MTLReadWriteTextureTier,
+    msaa_desktop: bool,
+    msaa_apple3: bool,
+    msaa_apple7: bool,
     resource_heaps: bool,
     argument_buffers: bool,
     shared_textures: bool,
     mutable_comparison_samplers: bool,
     sampler_clamp_to_border: bool,
-    sampler_lod_average: bool,
     base_instance: bool,
     base_vertex_instance_drawing: bool,
     dual_source_blending: bool,
@@ -167,6 +195,7 @@ struct PrivateCapabilities {
     format_bc: bool,
     format_eac_etc: bool,
     format_astc: bool,
+    format_astc_hdr: bool,
     format_any8_unorm_srgb_all: bool,
     format_any8_unorm_srgb_no_write: bool,
     format_any8_snorm_all: bool,
@@ -180,8 +209,7 @@ struct PrivateCapabilities {
     format_rgba8_srgb_no_write: bool,
     format_rgb10a2_unorm_all: bool,
     format_rgb10a2_unorm_no_write: bool,
-    format_rgb10a2_uint_color: bool,
-    format_rgb10a2_uint_color_write: bool,
+    format_rgb10a2_uint_write: bool,
     format_rg11b10_all: bool,
     format_rg11b10_no_write: bool,
     format_rgb9e5_all: bool,
@@ -203,6 +231,7 @@ struct PrivateCapabilities {
     format_bgr10a2_all: bool,
     format_bgr10a2_no_write: bool,
     max_buffers_per_stage: ResourceIndex,
+    max_vertex_buffers: ResourceIndex,
     max_textures_per_stage: ResourceIndex,
     max_samplers_per_stage: ResourceIndex,
     buffer_alignment: u64,
@@ -215,7 +244,7 @@ struct PrivateCapabilities {
     max_varying_components: u32,
     max_threads_per_group: u32,
     max_total_threadgroup_memory: u32,
-    sample_count_mask: u8,
+    sample_count_mask: crate::TextureFormatCapabilities,
     supports_debug_markers: bool,
     supports_binary_archives: bool,
     supports_capture_manager: bool,
@@ -227,6 +256,9 @@ struct PrivateCapabilities {
     supports_mutability: bool,
     supports_depth_clip_control: bool,
     supports_preserve_invariance: bool,
+    supports_shader_primitive_index: bool,
+    has_unified_memory: Option<bool>,
+    timestamp_query_support: TimestampQuerySupport,
 }
 
 #[derive(Clone, Debug)]
@@ -244,25 +276,27 @@ struct Settings {
 }
 
 struct AdapterShared {
-    device: Mutex<mtl::Device>,
+    device: Mutex<metal::Device>,
     disabilities: PrivateDisabilities,
     private_caps: PrivateCapabilities,
     settings: Settings,
+    presentation_timer: time::PresentationTimer,
 }
 
 unsafe impl Send for AdapterShared {}
 unsafe impl Sync for AdapterShared {}
 
 impl AdapterShared {
-    fn new(device: mtl::Device) -> Self {
+    fn new(device: metal::Device) -> Self {
         let private_caps = PrivateCapabilities::new(&device);
         log::debug!("{:#?}", private_caps);
 
         Self {
             disabilities: PrivateDisabilities::new(&device),
-            private_caps: PrivateCapabilities::new(&device),
+            private_caps,
             device: Mutex::new(device),
             settings: Settings::default(),
+            presentation_timer: time::PresentationTimer::new(),
         }
     }
 }
@@ -272,12 +306,22 @@ pub struct Adapter {
 }
 
 pub struct Queue {
-    raw: Arc<Mutex<mtl::CommandQueue>>,
+    raw: Arc<Mutex<metal::CommandQueue>>,
+    timestamp_period: f32,
 }
 
 unsafe impl Send for Queue {}
 unsafe impl Sync for Queue {}
 
+impl Queue {
+    pub unsafe fn queue_from_raw(raw: metal::CommandQueue, timestamp_period: f32) -> Self {
+        Self {
+            raw: Arc::new(Mutex::new(raw)),
+            timestamp_period,
+        }
+    }
+}
+
 pub struct Device {
     shared: Arc<AdapterShared>,
     features: wgt::Features,
@@ -285,8 +329,8 @@ pub struct Device {
 
 pub struct Surface {
     view: Option<NonNull<objc::runtime::Object>>,
-    render_layer: Mutex<mtl::MetalLayer>,
-    raw_swapchain_format: mtl::MTLPixelFormat,
+    render_layer: Mutex<metal::MetalLayer>,
+    swapchain_format: Option<wgt::TextureFormat>,
     extent: wgt::Extent3d,
     main_thread_id: thread::ThreadId,
     // Useful for UI-intensive applications that are sensitive to
@@ -300,7 +344,7 @@ unsafe impl Sync for Surface {}
 #[derive(Debug)]
 pub struct SurfaceTexture {
     texture: Texture,
-    drawable: mtl::MetalDrawable,
+    drawable: metal::MetalDrawable,
     present_with_transaction: bool,
 }
 
@@ -337,7 +381,7 @@ impl crate::Queue<Api> for Queue {
                                 .to_owned()
                         }
                     };
-                    raw.set_label("_Signal");
+                    raw.set_label("(wgpu internal) Signal");
                     raw.add_completed_handler(&block);
 
                     fence.maintain();
@@ -369,7 +413,7 @@ impl crate::Queue<Api> for Queue {
         let queue = &self.raw.lock();
         objc::rc::autoreleasepool(|| {
             let command_buffer = queue.new_command_buffer();
-            command_buffer.set_label("_Present");
+            command_buffer.set_label("(wgpu internal) Present");
 
             // https://developer.apple.com/documentation/quartzcore/cametallayer/1478157-presentswithtransaction?language=objc
             if !texture.present_with_transaction {
@@ -387,14 +431,13 @@ impl crate::Queue<Api> for Queue {
     }
 
     unsafe fn get_timestamp_period(&self) -> f32 {
-        // TODO: This is hard, see https://github.com/gpuweb/gpuweb/issues/1325
-        1.0
+        self.timestamp_period
     }
 }
 
 #[derive(Debug)]
 pub struct Buffer {
-    raw: mtl::Buffer,
+    raw: metal::Buffer,
     size: wgt::BufferAddress,
 }
 
@@ -409,9 +452,9 @@ impl Buffer {
 
 #[derive(Debug)]
 pub struct Texture {
-    raw: mtl::Texture,
-    raw_format: mtl::MTLPixelFormat,
-    raw_type: mtl::MTLTextureType,
+    raw: metal::Texture,
+    format: wgt::TextureFormat,
+    raw_type: metal::MTLTextureType,
     array_layers: u32,
     mip_levels: u32,
     copy_size: crate::CopyExtent,
@@ -422,7 +465,7 @@ unsafe impl Sync for Texture {}
 
 #[derive(Debug)]
 pub struct TextureView {
-    raw: mtl::Texture,
+    raw: metal::Texture,
     aspects: crate::FormatAspects,
 }
 
@@ -437,7 +480,7 @@ impl TextureView {
 
 #[derive(Debug)]
 pub struct Sampler {
-    raw: mtl::SamplerState,
+    raw: metal::SamplerState,
 }
 
 unsafe impl Send for Sampler {}
@@ -487,13 +530,20 @@ impl<T> ops::Index<naga::ShaderStage> for MultiStageData<T> {
 }
 
 impl<T> MultiStageData<T> {
-    fn map<Y>(&self, fun: impl Fn(&T) -> Y) -> MultiStageData<Y> {
+    fn map_ref<Y>(&self, fun: impl Fn(&T) -> Y) -> MultiStageData<Y> {
         MultiStageData {
             vs: fun(&self.vs),
             fs: fun(&self.fs),
             cs: fun(&self.cs),
         }
     }
+    fn map<Y>(self, fun: impl Fn(T) -> Y) -> MultiStageData<Y> {
+        MultiStageData {
+            vs: fun(self.vs),
+            fs: fun(self.fs),
+            cs: fun(self.cs),
+        }
+    }
     fn iter<'a>(&'a self) -> impl Iterator<Item = &'a T> {
         iter::once(&self.vs)
             .chain(iter::once(&self.fs))
@@ -507,6 +557,7 @@ impl<T> MultiStageData<T> {
 }
 
 type MultiStageResourceCounters = MultiStageData<ResourceData<ResourceIndex>>;
+type MultiStageResources = MultiStageData<naga::back::msl::EntryPointResources>;
 
 #[derive(Debug)]
 struct BindGroupLayoutInfo {
@@ -521,11 +572,11 @@ struct PushConstantsInfo {
 
 #[derive(Debug)]
 pub struct PipelineLayout {
-    naga_options: naga::back::msl::Options,
     bind_group_infos: ArrayVec<BindGroupLayoutInfo, { crate::MAX_BIND_GROUPS }>,
     push_constants_infos: MultiStageData<Option<PushConstantsInfo>>,
     total_counters: MultiStageResourceCounters,
     total_push_constants: u32,
+    per_stage_map: MultiStageResources,
 }
 
 trait AsNative {
@@ -534,12 +585,12 @@ trait AsNative {
     fn as_native(&self) -> &Self::Native;
 }
 
-type BufferPtr = NonNull<mtl::MTLBuffer>;
-type TexturePtr = NonNull<mtl::MTLTexture>;
-type SamplerPtr = NonNull<mtl::MTLSamplerState>;
+type BufferPtr = NonNull<metal::MTLBuffer>;
+type TexturePtr = NonNull<metal::MTLTexture>;
+type SamplerPtr = NonNull<metal::MTLSamplerState>;
 
 impl AsNative for BufferPtr {
-    type Native = mtl::BufferRef;
+    type Native = metal::BufferRef;
     #[inline]
     fn from(native: &Self::Native) -> Self {
         unsafe { NonNull::new_unchecked(native.as_ptr()) }
@@ -551,7 +602,7 @@ impl AsNative for BufferPtr {
 }
 
 impl AsNative for TexturePtr {
-    type Native = mtl::TextureRef;
+    type Native = metal::TextureRef;
     #[inline]
     fn from(native: &Self::Native) -> Self {
         unsafe { NonNull::new_unchecked(native.as_ptr()) }
@@ -563,7 +614,7 @@ impl AsNative for TexturePtr {
 }
 
 impl AsNative for SamplerPtr {
-    type Native = mtl::SamplerStateRef;
+    type Native = metal::SamplerStateRef;
     #[inline]
     fn from(native: &Self::Native) -> Self {
         unsafe { NonNull::new_unchecked(native.as_ptr()) }
@@ -579,7 +630,17 @@ struct BufferResource {
     ptr: BufferPtr,
     offset: wgt::BufferAddress,
     dynamic_index: Option<u32>,
+
+    /// The buffer's size, if it is a [`Storage`] binding. Otherwise `None`.
+    ///
+    /// Buffers with the [`wgt::BufferBindingType::Storage`] binding type can
+    /// hold WGSL runtime-sized arrays. When one does, we must pass its size to
+    /// shader entry points to implement bounds checks and WGSL's `arrayLength`
+    /// function. See [`device::CompiledShader::sized_bindings`] for details.
+    ///
+    /// [`Storage`]: wgt::BufferBindingType::Storage
     binding_size: Option<wgt::BufferSize>,
+
     binding_location: u32,
 }
 
@@ -597,12 +658,21 @@ unsafe impl Sync for BindGroup {}
 #[derive(Debug)]
 pub struct ShaderModule {
     naga: crate::NagaShader,
+    runtime_checks: bool,
 }
 
 #[derive(Debug, Default)]
 struct PipelineStageInfo {
     push_constants: Option<PushConstantsInfo>,
+
+    /// The buffer argument table index at which we pass runtime-sized arrays' buffer sizes.
+    ///
+    /// See [`device::CompiledShader::sized_bindings`] for more details.
     sizes_slot: Option<naga::back::msl::Slot>,
+
+    /// Bindings of all WGSL `storage` globals that contain runtime-sized arrays.
+    ///
+    /// See [`device::CompiledShader::sized_bindings`] for more details.
     sized_bindings: Vec<naga::ResourceBinding>,
 }
 
@@ -622,39 +692,41 @@ impl PipelineStageInfo {
 }
 
 pub struct RenderPipeline {
-    raw: mtl::RenderPipelineState,
+    raw: metal::RenderPipelineState,
     #[allow(dead_code)]
-    vs_lib: mtl::Library,
+    vs_lib: metal::Library,
     #[allow(dead_code)]
-    fs_lib: Option<mtl::Library>,
+    fs_lib: Option<metal::Library>,
     vs_info: PipelineStageInfo,
-    fs_info: PipelineStageInfo,
-    raw_primitive_type: mtl::MTLPrimitiveType,
-    raw_triangle_fill_mode: mtl::MTLTriangleFillMode,
-    raw_front_winding: mtl::MTLWinding,
-    raw_cull_mode: mtl::MTLCullMode,
-    raw_depth_clip_mode: Option<mtl::MTLDepthClipMode>,
-    depth_stencil: Option<(mtl::DepthStencilState, wgt::DepthBiasState)>,
+    fs_info: Option<PipelineStageInfo>,
+    raw_primitive_type: metal::MTLPrimitiveType,
+    raw_triangle_fill_mode: metal::MTLTriangleFillMode,
+    raw_front_winding: metal::MTLWinding,
+    raw_cull_mode: metal::MTLCullMode,
+    raw_depth_clip_mode: Option<metal::MTLDepthClipMode>,
+    depth_stencil: Option<(metal::DepthStencilState, wgt::DepthBiasState)>,
 }
 
 unsafe impl Send for RenderPipeline {}
 unsafe impl Sync for RenderPipeline {}
 
 pub struct ComputePipeline {
-    raw: mtl::ComputePipelineState,
+    raw: metal::ComputePipelineState,
     #[allow(dead_code)]
-    cs_lib: mtl::Library,
+    cs_lib: metal::Library,
     cs_info: PipelineStageInfo,
-    work_group_size: mtl::MTLSize,
+    work_group_size: metal::MTLSize,
     work_group_memory_sizes: Vec<u32>,
 }
 
 unsafe impl Send for ComputePipeline {}
 unsafe impl Sync for ComputePipeline {}
 
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 pub struct QuerySet {
-    raw_buffer: mtl::Buffer,
+    raw_buffer: metal::Buffer,
+    //Metal has a custom buffer for counters.
+    counter_sample_buffer: Option<metal::CounterSampleBuffer>,
     ty: wgt::QueryType,
 }
 
@@ -665,7 +737,7 @@ unsafe impl Sync for QuerySet {}
 pub struct Fence {
     completed_value: Arc<atomic::AtomicU64>,
     /// The pending fence values have to be ascending.
-    pending_command_buffers: Vec<(crate::FenceValue, mtl::CommandBuffer)>,
+    pending_command_buffers: Vec<(crate::FenceValue, metal::CommandBuffer)>,
 }
 
 unsafe impl Send for Fence {}
@@ -675,7 +747,7 @@ impl Fence {
     fn get_latest(&self) -> crate::FenceValue {
         let mut max_value = self.completed_value.load(atomic::Ordering::Acquire);
         for &(value, ref cmd_buf) in self.pending_command_buffers.iter() {
-            if cmd_buf.status() == mtl::MTLCommandBufferStatus::Completed {
+            if cmd_buf.status() == metal::MTLCommandBufferStatus::Completed {
                 max_value = value;
             }
         }
@@ -693,7 +765,7 @@ struct IndexState {
     buffer_ptr: BufferPtr,
     offset: wgt::BufferAddress,
     stride: wgt::BufferAddress,
-    raw_type: mtl::MTLIndexType,
+    raw_type: metal::MTLIndexType,
 }
 
 #[derive(Default)]
@@ -702,31 +774,65 @@ struct Temp {
 }
 
 struct CommandState {
-    blit: Option<mtl::BlitCommandEncoder>,
-    render: Option<mtl::RenderCommandEncoder>,
-    compute: Option<mtl::ComputeCommandEncoder>,
-    raw_primitive_type: mtl::MTLPrimitiveType,
+    blit: Option<metal::BlitCommandEncoder>,
+    render: Option<metal::RenderCommandEncoder>,
+    compute: Option<metal::ComputeCommandEncoder>,
+    raw_primitive_type: metal::MTLPrimitiveType,
     index: Option<IndexState>,
-    raw_wg_size: mtl::MTLSize,
+    raw_wg_size: metal::MTLSize,
     stage_infos: MultiStageData<PipelineStageInfo>,
-    storage_buffer_length_map: fxhash::FxHashMap<naga::ResourceBinding, wgt::BufferSize>,
+
+    /// Sizes of currently bound [`wgt::BufferBindingType::Storage`] buffers.
+    ///
+    /// Specifically:
+    ///
+    /// - The keys are ['ResourceBinding`] values (that is, the WGSL `@group`
+    ///   and `@binding` attributes) for `var<storage>` global variables in the
+    ///   current module that contain runtime-sized arrays.
+    ///
+    /// - The values are the actual sizes of the buffers currently bound to
+    ///   provide those globals' contents, which are needed to implement bounds
+    ///   checks and the WGSL `arrayLength` function.
+    ///
+    /// For each stage `S` in `stage_infos`, we consult this to find the sizes
+    /// of the buffers listed in [`stage_infos.S.sized_bindings`], which we must
+    /// pass to the entry point.
+    ///
+    /// See [`device::CompiledShader::sized_bindings`] for more details.
+    ///
+    /// [`ResourceBinding`]: naga::ResourceBinding
+    storage_buffer_length_map: rustc_hash::FxHashMap<naga::ResourceBinding, wgt::BufferSize>,
+
     work_group_memory_sizes: Vec<u32>,
     push_constants: Vec<u32>,
+
+    /// Timer query that should be executed when the next pass starts.
+    pending_timer_queries: Vec<(QuerySet, u32)>,
 }
 
 pub struct CommandEncoder {
     shared: Arc<AdapterShared>,
-    raw_queue: Arc<Mutex<mtl::CommandQueue>>,
-    raw_cmd_buf: Option<mtl::CommandBuffer>,
+    raw_queue: Arc<Mutex<metal::CommandQueue>>,
+    raw_cmd_buf: Option<metal::CommandBuffer>,
     state: CommandState,
     temp: Temp,
 }
 
+impl fmt::Debug for CommandEncoder {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("CommandEncoder")
+            .field("raw_queue", &self.raw_queue)
+            .field("raw_cmd_buf", &self.raw_cmd_buf)
+            .finish()
+    }
+}
+
 unsafe impl Send for CommandEncoder {}
 unsafe impl Sync for CommandEncoder {}
 
+#[derive(Debug)]
 pub struct CommandBuffer {
-    raw: mtl::CommandBuffer,
+    raw: metal::CommandBuffer,
 }
 
 unsafe impl Send for CommandBuffer {}
