@@ -1,3 +1,5 @@
+#![allow(clippy::let_unit_value)] // `let () =` being used to constrain result type
+
 use std::{mem, os::raw::c_void, ptr::NonNull, sync::Once, thread};
 
 use core_graphics_types::{
@@ -9,13 +11,13 @@ use objc::{
     declare::ClassDecl,
     msg_send,
     rc::autoreleasepool,
-    runtime::{Class, Object, Sel, BOOL, YES},
+    runtime::{Class, Object, Sel, BOOL, NO, YES},
     sel, sel_impl,
 };
 use parking_lot::Mutex;
 
 #[cfg(target_os = "macos")]
-#[link(name = "QuartzCore", kind = "framework")]
+#[cfg_attr(feature = "link", link(name = "QuartzCore", kind = "framework"))]
 extern "C" {
     #[allow(non_upper_case_globals)]
     static kCAGravityTopLeft: *mut Object;
@@ -57,12 +59,11 @@ impl HalManagedMetalLayerDelegate {
 }
 
 impl super::Surface {
-    fn new(view: Option<NonNull<Object>>, layer: mtl::MetalLayer) -> Self {
+    fn new(view: Option<NonNull<Object>>, layer: metal::MetalLayer) -> Self {
         Self {
             view,
             render_layer: Mutex::new(layer),
-            swapchain_format: wgt::TextureFormat::Bgra8UnormSrgb, // no value invalid, pick something not too far-fetched
-            raw_swapchain_format: mtl::MTLPixelFormat::Invalid,
+            swapchain_format: None,
             extent: wgt::Extent3d::default(),
             main_thread_id: thread::current().id(),
             present_with_transaction: false,
@@ -75,124 +76,89 @@ impl super::Surface {
         }
     }
 
-    #[cfg(target_os = "ios")]
+    /// If not called on the main thread, this will panic.
     #[allow(clippy::transmute_ptr_to_ref)]
-    pub unsafe fn from_uiview(uiview: *mut c_void) -> Self {
-        let view = uiview as *mut Object;
-        if view.is_null() {
-            panic!("window does not have a valid contentView");
-        }
-
-        let main_layer: *mut Object = msg_send![view, layer];
-        let class = class!(CAMetalLayer);
-        let is_valid_layer: BOOL = msg_send![main_layer, isKindOfClass: class];
-        let render_layer = if is_valid_layer == YES {
-            mem::transmute::<_, &mtl::MetalLayerRef>(main_layer).to_owned()
-        } else {
-            // If the main layer is not a CAMetalLayer, we create a CAMetalLayer sublayer and use it instead.
-            // Unlike on macOS, we cannot replace the main view as UIView does not allow it (when NSView does).
-            let new_layer: mtl::MetalLayer = msg_send![class, new];
-            let bounds: CGRect = msg_send![main_layer, bounds];
-            let () = msg_send![new_layer.as_ref(), setFrame: bounds];
-            let () = msg_send![main_layer, addSublayer: new_layer.as_ref()];
-            new_layer
-        };
-
-        let window: *mut Object = msg_send![view, window];
-        if !window.is_null() {
-            let screen: *mut Object = msg_send![window, screen];
-            assert!(!screen.is_null(), "window is not attached to a screen");
-
-            let scale_factor: CGFloat = msg_send![screen, nativeScale];
-            let () = msg_send![view, setContentScaleFactor: scale_factor];
-        }
-
-        let _: *mut c_void = msg_send![view, retain];
-        Self::new(NonNull::new(view), render_layer)
-    }
-
-    #[cfg(target_os = "macos")]
-    #[allow(clippy::transmute_ptr_to_ref)]
-    pub unsafe fn from_nsview(
-        nsview: *mut c_void,
-        delegate: &HalManagedMetalLayerDelegate,
+    pub unsafe fn from_view(
+        view: *mut c_void,
+        delegate: Option<&HalManagedMetalLayerDelegate>,
     ) -> Self {
-        let view = nsview as *mut Object;
-        if view.is_null() {
-            panic!("window does not have a valid contentView");
+        let view = view as *mut Object;
+        let render_layer = {
+            let layer = unsafe { Self::get_metal_layer(view, delegate) };
+            unsafe { mem::transmute::<_, &metal::MetalLayerRef>(layer) }
         }
-
-        let class = class!(CAMetalLayer);
-        // Deprecated! Clients should use `create_surface_from_layer` instead.
-        let is_actually_layer: BOOL = msg_send![view, isKindOfClass: class];
-        if is_actually_layer == YES {
-            return Self::from_layer(mem::transmute(view));
-        }
-
-        let existing: *mut Object = msg_send![view, layer];
-        let use_current = if existing.is_null() {
-            false
-        } else {
-            let result: BOOL = msg_send![existing, isKindOfClass: class];
-            result == YES
-        };
-
-        let render_layer: mtl::MetalLayer = if use_current {
-            mem::transmute::<_, &mtl::MetalLayerRef>(existing).to_owned()
-        } else {
-            let layer: mtl::MetalLayer = msg_send![class, new];
-            let () = msg_send![view, setLayer: layer.as_ref()];
-            let () = msg_send![view, setWantsLayer: YES];
-            let bounds: CGRect = msg_send![view, bounds];
-            let () = msg_send![layer.as_ref(), setBounds: bounds];
-
-            let window: *mut Object = msg_send![view, window];
-            if !window.is_null() {
-                let scale_factor: CGFloat = msg_send![window, backingScaleFactor];
-                let () = msg_send![layer, setContentsScale: scale_factor];
-            }
-            let () = msg_send![layer, setDelegate: delegate.0];
-            layer
-        };
-
-        let () = msg_send![render_layer, setContentsGravity: kCAGravityTopLeft];
-
+        .to_owned();
         let _: *mut c_void = msg_send![view, retain];
         Self::new(NonNull::new(view), render_layer)
     }
 
-    pub unsafe fn from_layer(layer: &mtl::MetalLayerRef) -> Self {
+    pub unsafe fn from_layer(layer: &metal::MetalLayerRef) -> Self {
         let class = class!(CAMetalLayer);
         let proper_kind: BOOL = msg_send![layer, isKindOfClass: class];
         assert_eq!(proper_kind, YES);
         Self::new(None, layer.to_owned())
     }
 
-    pub(super) fn dimensions(&self) -> wgt::Extent3d {
-        let (size, scale): (CGSize, CGFloat) = match self.view {
-            Some(view) if !cfg!(target_os = "macos") => unsafe {
-                let bounds: CGRect = msg_send![view.as_ptr(), bounds];
-                let window: Option<NonNull<Object>> = msg_send![view.as_ptr(), window];
-                let screen = window.and_then(|window| -> Option<NonNull<Object>> {
-                    msg_send![window.as_ptr(), screen]
-                });
-                match screen {
-                    Some(screen) => {
-                        let screen_space: *mut Object = msg_send![screen.as_ptr(), coordinateSpace];
-                        let rect: CGRect = msg_send![view.as_ptr(), convertRect:bounds toCoordinateSpace:screen_space];
-                        let scale_factor: CGFloat = msg_send![screen.as_ptr(), nativeScale];
-                        (rect.size, scale_factor)
-                    }
-                    None => (bounds.size, 1.0),
+    /// If not called on the main thread, this will panic.
+    pub(crate) unsafe fn get_metal_layer(
+        view: *mut Object,
+        delegate: Option<&HalManagedMetalLayerDelegate>,
+    ) -> *mut Object {
+        if view.is_null() {
+            panic!("window does not have a valid contentView");
+        }
+
+        let is_main_thread: BOOL = msg_send![class!(NSThread), isMainThread];
+        if is_main_thread == NO {
+            panic!("get_metal_layer cannot be called in non-ui thread.");
+        }
+
+        let main_layer: *mut Object = msg_send![view, layer];
+        let class = class!(CAMetalLayer);
+        let is_valid_layer: BOOL = msg_send![main_layer, isKindOfClass: class];
+
+        if is_valid_layer == YES {
+            main_layer
+        } else {
+            // If the main layer is not a CAMetalLayer, we create a CAMetalLayer and use it.
+            let new_layer: *mut Object = msg_send![class, new];
+            let frame: CGRect = msg_send![main_layer, bounds];
+            let () = msg_send![new_layer, setFrame: frame];
+            #[cfg(target_os = "ios")]
+            {
+                // Unlike NSView, UIView does not allow to replace main layer.
+                let () = msg_send![main_layer, addSublayer: new_layer];
+                // On iOS, "from_view" may be called before the application initialization is complete,
+                // `msg_send![view, window]` and `msg_send![window, screen]` will get null.
+                let screen: *mut Object = msg_send![class!(UIScreen), mainScreen];
+                let scale_factor: CGFloat = msg_send![screen, nativeScale];
+                let () = msg_send![view, setContentScaleFactor: scale_factor];
+            };
+            #[cfg(target_os = "macos")]
+            {
+                let () = msg_send![view, setLayer: new_layer];
+                let () = msg_send![view, setWantsLayer: YES];
+                let () = msg_send![new_layer, setContentsGravity: unsafe { kCAGravityTopLeft }];
+                let window: *mut Object = msg_send![view, window];
+                if !window.is_null() {
+                    let scale_factor: CGFloat = msg_send![window, backingScaleFactor];
+                    let () = msg_send![new_layer, setContentsScale: scale_factor];
                 }
-            },
-            _ => unsafe {
-                let render_layer_borrow = self.render_layer.lock();
-                let render_layer = render_layer_borrow.as_ref();
-                let bounds: CGRect = msg_send![render_layer, bounds];
-                let contents_scale: CGFloat = msg_send![render_layer, contentsScale];
-                (bounds.size, contents_scale)
-            },
+            };
+            if let Some(delegate) = delegate {
+                let () = msg_send![new_layer, setDelegate: delegate.0];
+            }
+            new_layer
+        }
+    }
+
+    pub(super) fn dimensions(&self) -> wgt::Extent3d {
+        let (size, scale): (CGSize, CGFloat) = unsafe {
+            let render_layer_borrow = self.render_layer.lock();
+            let render_layer = render_layer_borrow.as_ref();
+            let bounds: CGRect = msg_send![render_layer, bounds];
+            let contents_scale: CGFloat = msg_send![render_layer, contentsScale];
+            (bounds.size, contents_scale)
         };
 
         wgt::Extent3d {
@@ -212,19 +178,22 @@ impl crate::Surface<super::Api> for super::Surface {
         log::info!("build swapchain {:?}", config);
 
         let caps = &device.shared.private_caps;
-        self.swapchain_format = config.format;
-        self.raw_swapchain_format = caps.map_format(config.format);
+        self.swapchain_format = Some(config.format);
         self.extent = config.extent;
 
         let render_layer = self.render_layer.lock();
         let framebuffer_only = config.usage == crate::TextureUses::COLOR_TARGET;
-        let display_sync = config.present_mode != wgt::PresentMode::Immediate;
+        let display_sync = match config.present_mode {
+            wgt::PresentMode::Fifo => true,
+            wgt::PresentMode::Immediate => false,
+            m => unreachable!("Unsupported present mode: {m:?}"),
+        };
         let drawable_size = CGSize::new(config.extent.width as f64, config.extent.height as f64);
 
         match config.composite_alpha_mode {
-            crate::CompositeAlphaMode::Opaque => render_layer.set_opaque(true),
-            crate::CompositeAlphaMode::PostMultiplied => render_layer.set_opaque(false),
-            crate::CompositeAlphaMode::PreMultiplied => (),
+            wgt::CompositeAlphaMode::Opaque => render_layer.set_opaque(true),
+            wgt::CompositeAlphaMode::PostMultiplied => render_layer.set_opaque(false),
+            _ => (),
         }
 
         let device_raw = device.shared.device.lock();
@@ -240,14 +209,19 @@ impl crate::Surface<super::Api> for super::Surface {
                 let () = msg_send![*render_layer, setFrame: bounds];
             }
         }
-        render_layer.set_device(&*device_raw);
-        render_layer.set_pixel_format(self.raw_swapchain_format);
+        render_layer.set_device(&device_raw);
+        render_layer.set_pixel_format(caps.map_format(config.format));
         render_layer.set_framebuffer_only(framebuffer_only);
         render_layer.set_presents_with_transaction(self.present_with_transaction);
+        // opt-in to Metal EDR
+        // EDR potentially more power used in display and more bandwidth, memory footprint.
+        let wants_edr = config.format == wgt::TextureFormat::Rgba16Float;
+        if wants_edr != render_layer.wants_extended_dynamic_range_content() {
+            render_layer.set_wants_extended_dynamic_range_content(wants_edr);
+        }
 
         // this gets ignored on iOS for certain OS/device combinations (iphone5s iOS 10.3)
-        let () = msg_send![*render_layer, setMaximumDrawableCount: config.swap_chain_size as u64];
-
+        render_layer.set_maximum_drawable_count(config.swap_chain_size as _);
         render_layer.set_drawable_size(drawable_size);
         if caps.can_set_next_drawable_timeout {
             let () = msg_send![*render_layer, setAllowsNextDrawableTimeout:false];
@@ -260,12 +234,12 @@ impl crate::Surface<super::Api> for super::Surface {
     }
 
     unsafe fn unconfigure(&mut self, _device: &super::Device) {
-        self.raw_swapchain_format = mtl::MTLPixelFormat::Invalid;
+        self.swapchain_format = None;
     }
 
     unsafe fn acquire_texture(
         &mut self,
-        _timeout_ms: u32, //TODO
+        _timeout_ms: Option<std::time::Duration>, //TODO
     ) -> Result<Option<crate::AcquiredSurfaceTexture<super::Api>>, crate::SurfaceError> {
         let render_layer = self.render_layer.lock();
         let (drawable, texture) = match autoreleasepool(|| {
@@ -280,9 +254,8 @@ impl crate::Surface<super::Api> for super::Surface {
         let suf_texture = super::SurfaceTexture {
             texture: super::Texture {
                 raw: texture,
-                format: self.swapchain_format,
-                raw_format: self.raw_swapchain_format,
-                raw_type: mtl::MTLTextureType::D2,
+                format: self.swapchain_format.unwrap(),
+                raw_type: metal::MTLTextureType::D2,
                 array_layers: 1,
                 mip_levels: 1,
                 copy_size: crate::CopyExtent {
